@@ -29,7 +29,10 @@ use nextest_runner::{
     },
     partition::PartitionerBuilder,
     platform::BuildPlatforms,
-    reporter::{FinalStatusLevel, ReporterBuilder, ShowProgress, StatusLevel, TestOutputDisplay},
+    reporter::{
+        FinalStatusLevel, MaxProgressRunning, ReporterBuilder, ShowProgress, StatusLevel,
+        TestOutputDisplay,
+    },
     reuse_build::ReuseBuildInfo,
     runner::{StressCondition, StressCount, TestRunnerBuilder},
     test_filter::{FilterBound, RunIgnored, TestFilterBuilder, TestFilterPatterns},
@@ -657,11 +660,17 @@ pub struct TestRunnerOpts {
     )]
     no_fail_fast: bool,
 
-    /// Number of tests that can fail before exiting test run [possible values: integer or "all"]
+    /// Number of tests that can fail before exiting test run
+    ///
+    /// To control whether currently running tests are waited for or terminated
+    /// immediately, append ':wait' (default) or ':immediate' to the number
+    /// (e.g., '5:immediate').
+    ///
+    /// [possible values: integer, "all", "N:wait", "N:immediate"]
     #[arg(
         long,
         name = "max-fail",
-        value_name = "N",
+        value_name = "N[:MODE]",
         conflicts_with_all = &["no-run", "fail-fast", "no-fail-fast"],
     )]
     max_fail: Option<MaxFail>,
@@ -865,6 +874,21 @@ pub(super) struct ReporterOpts {
     #[arg(long, env = "NEXTEST_NO_INPUT_HANDLER", value_parser = BoolishValueParser::new())]
     pub(super) no_input_handler: bool,
 
+    /// Maximum number of running tests to display progress for.
+    ///
+    /// When more tests are running than this limit, the progress bar will show
+    /// the first N tests and a summary of remaining tests (e.g. "... and 24
+    /// more tests running"). Set to **0** to hide running tests, or
+    /// **infinite** for unlimited. This applies when using
+    /// `--show-progress=bar` or `--show-progress=only`.
+    #[arg(
+        long = "max-progress-running",
+        value_name = "N",
+        env = "NEXTEST_MAX_PROGRESS_RUNNING",
+        default_value = "8"
+    )]
+    max_progress_running: MaxProgressRunning,
+
     /// Format to use for test results (experimental).
     #[arg(
         long,
@@ -964,6 +988,7 @@ impl ReporterOpts {
         }
         builder.set_show_progress(show_progress.into());
         builder.set_no_output_indent(self.no_output_indent);
+        builder.set_max_progress_running(self.max_progress_running);
         builder
     }
 }
@@ -1050,16 +1075,15 @@ enum ShowProgressOpt {
     /// Do not display a progress bar or counter.
     None,
 
-    /// Display a progress bar: default for interactive terminals.
+    /// Display a progress bar with running tests: default for interactive
+    /// terminals.
+    #[clap(alias = "running")]
     Bar,
 
     /// Display a counter next to each completed test.
     Counter,
 
-    /// Display each running test on a separate line.
-    Running,
-
-    /// Display each running test on a separate line, and hide successful test
+    /// Display a progress bar with running tests, and hide successful test
     /// output; equivalent to `--show-progress=running --status-level=slow
     /// --final-status-level=none`.
     Only,
@@ -1070,9 +1094,8 @@ impl From<ShowProgressOpt> for ShowProgress {
         match opt {
             ShowProgressOpt::Auto => ShowProgress::Auto,
             ShowProgressOpt::None => ShowProgress::None,
-            ShowProgressOpt::Bar => ShowProgress::Bar,
+            ShowProgressOpt::Bar => ShowProgress::Running,
             ShowProgressOpt::Counter => ShowProgress::Counter,
-            ShowProgressOpt::Running => ShowProgress::Running,
             ShowProgressOpt::Only => ShowProgress::Running,
         }
     }
