@@ -3,23 +3,87 @@
 
 use crate::{
     cargo_config::{CargoConfigs, DiscoveredConfig},
-    helpers::DisplayTestInstance,
+    helpers::{DisplayTestInstance, plural},
     list::TestInstanceId,
-    reporter::{displayer::formatters::DisplayBracketedHhMmSs, events::*, helpers::Styles},
+    reporter::{
+        displayer::formatters::DisplayBracketedHhMmSs,
+        events::*,
+        helpers::{Styles, print_lines_in_chunks},
+    },
 };
-use iddqd::{IdHashItem, IdHashMap, id_upcast};
-use indicatif::{MultiProgress, ProgressBar, ProgressDrawTarget, ProgressStyle};
+use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
 use nextest_metadata::RustBinaryId;
 use owo_colors::OwoColorize;
 use std::{
-    collections::VecDeque,
+    cmp::{max, min},
     env, fmt,
-    io::{self, IsTerminal, Write},
-    time::Duration,
+    io::IsTerminal,
+    str::FromStr,
+    time::{Duration, Instant},
 };
 use swrite::{SWrite, swrite};
 use tracing::debug;
 
+/// The refresh rate for the progress bar, set to a minimal value.
+///
+/// For progress, during each tick, two things happen:
+///
+/// - We update the message, calling self.bar.set_message.
+/// - We print any buffered output.
+///
+/// We want both of these updates to be combined into one terminal flush, so we
+/// set *this* to a minimal value (so self.bar.set_message doesn't do a redraw),
+/// and rely on ProgressBar::print_and_flush_buffer to always flush the
+/// terminal.
+const PROGRESS_REFRESH_RATE_HZ: u8 = 1;
+
+/// The maximum number of running tests to display with
+/// `--show-progress=running` or `only`.
+#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+pub enum MaxProgressRunning {
+    /// Show a specific maximum number of running tests.
+    /// If 0, running tests (including the overflow summary) aren't displayed.
+    Count(usize),
+
+    /// Show all running tests (no limit).
+    Infinite,
+}
+
+impl MaxProgressRunning {
+    /// The default value (8 tests).
+    pub const DEFAULT_VALUE: Self = Self::Count(8);
+}
+
+impl Default for MaxProgressRunning {
+    fn default() -> Self {
+        Self::DEFAULT_VALUE
+    }
+}
+
+impl FromStr for MaxProgressRunning {
+    type Err = String;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        if s.eq_ignore_ascii_case("infinite") {
+            return Ok(Self::Infinite);
+        }
+
+        match s.parse::<usize>() {
+            Err(e) => Err(format!("Error: {e} parsing {s}")),
+            Ok(n) => Ok(Self::Count(n)),
+        }
+    }
+}
+
+impl fmt::Display for MaxProgressRunning {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::Infinite => write!(f, "infinite"),
+            Self::Count(n) => write!(f, "{n}"),
+        }
+    }
+}
+
 /// How to show progress.
 #[derive(Default, Clone, Copy, Debug)]
 pub enum ShowProgress {
@@ -30,9 +94,6 @@ pub enum ShowProgress {
     /// No progress display.
     None,
 
-    /// Show a progress bar.
-    Bar,
-
     /// Show a counter on each line.
     Counter,
 
@@ -41,176 +102,72 @@ pub enum ShowProgress {
 }
 
 #[derive(Debug)]
-pub(super) struct TestProgressBars {
-    // A list of currently unused progress bars.
-    //
-    // Once a test bar is removed, we add an empty bar, then use it to avoid the
-    // overall progress bar shifting rapidly up and down.
-    free_bars: VecDeque<ProgressBar>,
-    used_bars: IdHashMap<TestProgressBar>,
-    spinner_chars: &'static str,
-}
-
-impl TestProgressBars {
-    fn new(spinner_chars: &'static str) -> Self {
-        Self {
-            free_bars: VecDeque::new(),
-            used_bars: IdHashMap::new(),
-            spinner_chars,
-        }
-    }
-
-    /// Adds a test to the bar if it's not present, or reset its duration to zero
-    /// if it is.
-    fn add_test(
-        &mut self,
-        multi: &MultiProgress,
-        id: &TestInstanceId<'_>,
-        prefix: String,
-        styles: &Styles,
-    ) {
-        if let Some(tb) = self.used_bars.get(id) {
-            // Reset the bar to 0.
-            tb.bar.set_prefix(prefix);
-            // `ProgressBar` is a lightweight handle and calling methods like
-            // `with_elapsed` on a clone also affects the original. Ideally
-            // there would be a `set_elapsed` method on self.bar, though.
-            tb.bar.clone().with_elapsed(Duration::ZERO);
-        } else {
-            self.add_test_inner(multi, id, prefix, styles);
-        }
-    }
-
-    /// Sets the prefix for a test bar.
-    fn set_prefix(&mut self, id: &TestInstanceId<'_>, prefix: String) {
-        if let Some(tb) = self.used_bars.get(id) {
-            tb.bar.set_prefix(prefix);
-        }
-    }
-
-    fn remove_test(&mut self, multi: &MultiProgress, id: &TestInstanceId<'_>) {
-        if let Some(data) = self.used_bars.remove(id) {
-            // Remove the progress bar from the multi-progress, and add a free
-            // bar at the end. This avoids output being rapidly shifted up and
-            // down.
-            multi.remove(&data.bar);
-            // Add an empty bar.
-            let free_bar = ProgressBar::hidden();
-            free_bar.set_style(
-                ProgressStyle::default_spinner()
-                    // Use a blank bar for the empty lines at the end -- a
-                    // minimum of one character is needed to force indicatif to
-                    // render this bar.
-                    .template(" ")
-                    .expect("this template is valid"),
-            );
-            let free_bar = multi.add(free_bar);
-            // Do an initial tick to update the display.
-            free_bar.tick();
-            self.free_bars.push_back(free_bar);
-        }
-    }
-
-    fn on_run_continued(&self, delta: Duration) {
-        for tb in &self.used_bars {
-            let current_elapsed = tb.bar.elapsed();
-            // `ProgressBar` is a lightweight handle and calling methods like
-            // `with_elapsed` on a clone also affects the original. Ideally
-            // there would be a `set_elapsed` method on self.bar, though.
-            tb.bar
-                .clone()
-                .with_elapsed(current_elapsed.saturating_sub(delta));
-        }
-    }
-
-    /// Adds a test to the progress display.
-    fn add_test_inner(
-        &mut self,
-        multi: &MultiProgress,
-        id: &TestInstanceId<'_>,
-        prefix: String,
-        styles: &Styles,
-    ) {
-        let free_bar = self.free_bars.pop_front();
-        let new_bar = ProgressBar::hidden();
-
-        new_bar.set_style(
-            ProgressStyle::default_bar()
-                .template("{prefix:>10} {spinner} [{elapsed_precise:>9}] {wide_msg}")
-                .expect("template to be valid")
-                .tick_chars(self.spinner_chars),
-        );
-        if !prefix.is_empty() {
-            new_bar.set_prefix(prefix);
-        }
-        new_bar.set_message(
-            DisplayTestInstance::new(None, None, *id, &styles.list_styles).to_string(),
-        );
-
-        // Remove the free bar from the multi-progress, then immediately insert
-        // the new bar into it. The free bar has done its job of providing
-        // padding.
-        if let Some(bar) = free_bar {
-            multi.remove(&bar);
-        }
-        // Insert at index self.used_bars.len() + 1 (the first bar is overall
-        // progress).
-        let bar = multi.insert(self.used_bars.len() + 1, new_bar);
-        // Do an initial tick after adding to the multi-progress to ensure
-        // that the new test is rendered.
-        bar.tick();
-
-        // Add the bar to the map.
-        self.used_bars.insert_overwrite(TestProgressBar {
-            binary_id: id.binary_id.clone(),
-            test_name: id.test_name.to_owned(),
-            bar,
-        });
-    }
-
-    fn tick_all(&self) {
-        // XXX tick in sequence rather than random order?
-        for tb in &self.used_bars {
-            tb.bar.tick();
-        }
-    }
-
-    fn finish_and_clear_all(&self) {
-        // XXX finish in sequence rather than random order?
-        for tb in &self.used_bars {
-            tb.bar.finish_and_clear();
-        }
-        for bar in &self.free_bars {
-            bar.finish_and_clear();
-        }
-    }
+pub(super) enum RunningTestStatus {
+    Running,
+    Slow,
+    Delay(Duration),
+    Retry,
 }
 
 #[derive(Debug)]
-pub(super) struct TestProgressBar {
+pub(super) struct RunningTest {
     binary_id: RustBinaryId,
     test_name: String,
-    bar: ProgressBar,
+    status: RunningTestStatus,
+    start_time: Instant,
+    paused_for: Duration,
 }
 
-impl IdHashItem for TestProgressBar {
-    type Key<'a> = TestInstanceId<'a>;
+impl RunningTest {
+    fn message(&self, now: &Instant, width: usize, styles: &Styles) -> String {
+        let mut elapsed = (*now - self.start_time) - self.paused_for;
+        let status = match self.status {
+            RunningTestStatus::Running => "     ".to_owned(),
+            RunningTestStatus::Slow => " SLOW".style(styles.skip).to_string(),
+            RunningTestStatus::Delay(d) => {
+                elapsed = d - elapsed;
+                "DELAY".style(styles.retry).to_string()
+            }
+            RunningTestStatus::Retry => "RETRY".style(styles.retry).to_string(),
+        };
+        let elapsed = format!(
+            "{:0>2}:{:0>2}:{:0>2}",
+            elapsed.as_secs() / 3600,
+            elapsed.as_secs() / 60,
+            elapsed.as_secs() % 60,
+        );
+        let max_width = width.saturating_sub(25);
+        let test = DisplayTestInstance::new(
+            None,
+            None,
+            TestInstanceId {
+                binary_id: &self.binary_id,
 
-    fn key(&self) -> Self::Key<'_> {
-        TestInstanceId {
-            binary_id: &self.binary_id,
-            test_name: &self.test_name,
-        }
+                test_name: &self.test_name,
+            },
+            &styles.list_styles,
+        )
+        .with_max_width(max_width);
+        format!("       {} [{:>9}] {}", status, elapsed, test)
     }
-
-    id_upcast!();
 }
 
 #[derive(Debug)]
 pub(super) struct ProgressBarState {
-    multi_progress: MultiProgress,
     bar: ProgressBar,
-    test_bars: Option<TestProgressBars>,
+    stats: RunStats,
+    running: usize,
+    max_progress_running: MaxProgressRunning,
+    // Keep track of the maximum number of lines used. This allows to adapt the
+    // size of the 'viewport' to what we are using, and not just to the maximum
+    // number of tests that can be run in parallel
+    max_running_displayed: usize,
+    // None when the running tests are not displayed
+    running_tests: Option<Vec<RunningTest>>,
+    buffer: String,
+    // Size in bytes for chunking println calls. Configurable via the
+    // undocumented __NEXTEST_PROGRESS_PRINTLN_CHUNK_SIZE env var.
+    println_chunk_size: usize,
     // Reasons for hiding the progress bar. We show the progress bar if none of
     // these are set and hide it if any of them are set.
     //
@@ -230,14 +187,9 @@ impl ProgressBarState {
     pub(super) fn new(
         test_count: usize,
         progress_chars: &str,
-        spinner_chars: &'static str,
-        show_running: bool,
+        max_progress_running: MaxProgressRunning,
     ) -> Self {
-        let multi_progress = MultiProgress::new();
-        multi_progress.set_draw_target(Self::stderr_target());
-        multi_progress.set_move_cursor(true);
-
-        let bar = multi_progress.add(ProgressBar::new(test_count as u64));
+        let bar = ProgressBar::new(test_count as u64);
         let test_count_width = format!("{test_count}").len();
         // Create the template using the width as input. This is a
         // little confusing -- {{foo}} is what's passed into the
@@ -245,7 +197,7 @@ impl ProgressBarState {
         // statement.
         let template = format!(
             "{{prefix:>12}} [{{elapsed_precise:>9}}] {{wide_bar}} \
-            {{pos:>{test_count_width}}}/{{len:{test_count_width}}}: {{msg}}     "
+            {{pos:>{test_count_width}}}/{{len:{test_count_width}}}: {{msg}}"
         );
         bar.set_style(
             ProgressStyle::default_bar()
@@ -254,12 +206,26 @@ impl ProgressBarState {
                 .expect("template is known to be valid"),
         );
 
-        let test_bars = show_running.then(|| TestProgressBars::new(spinner_chars));
+        let running_tests =
+            (!matches!(max_progress_running, MaxProgressRunning::Count(0))).then(Vec::new);
+
+        // The println chunk size defaults to a value chosen by experimentation,
+        // locally and over SSH. This controls how often the progress bar
+        // refreshes during large output bursts.
+        let println_chunk_size = env::var("__NEXTEST_PROGRESS_PRINTLN_CHUNK_SIZE")
+            .ok()
+            .and_then(|s| s.parse::<usize>().ok())
+            .unwrap_or(4096);
 
         Self {
-            multi_progress,
             bar,
-            test_bars,
+            stats: RunStats::default(),
+            running: 0,
+            max_progress_running,
+            max_running_displayed: 0,
+            running_tests,
+            buffer: String::new(),
+            println_chunk_size,
             hidden_no_capture: false,
             hidden_run_paused: false,
             hidden_info_response: false,
@@ -267,24 +233,85 @@ impl ProgressBarState {
         }
     }
 
-    pub(super) fn tick(&mut self) {
-        self.bar.tick();
-        // Also tick all test bars.
-        if let Some(test_bars) = &mut self.test_bars {
-            test_bars.tick_all();
+    pub(super) fn tick(&mut self, styles: &Styles) {
+        self.update_message(styles);
+        self.print_and_clear_buffer();
+    }
+
+    fn print_and_clear_buffer(&mut self) {
+        self.print_and_force_redraw();
+        self.buffer.clear();
+    }
+
+    /// Prints the contents of the buffer, and always forces a redraw.
+    fn print_and_force_redraw(&self) {
+        if self.buffer.is_empty() {
+            // Force a redraw as part of our contract. See the documentation for
+            // `PROGRESS_REFRESH_RATE_HZ`.
+            self.bar.force_draw();
+            return;
         }
+
+        // println below also forces a redraw, so we don't need to call
+        // force_draw in this case.
+
+        // ProgressBar::println is only called if there's something in the
+        // buffer, for two reasons:
+        //
+        // 1. If passed in nothing at all, it prints an empty line.
+        // 2. It forces a full redraw.
+        //
+        // But if self.buffer is too large, we can overwhelm the terminal with
+        // large amounts of non-progress-bar output, causing the progress bar to
+        // flicker in and out. To avoid those issues, we chunk the output to
+        // maintain progress bar visibility by redrawing it regularly.
+        print_lines_in_chunks(&self.buffer, self.println_chunk_size, |chunk| {
+            self.bar.println(chunk);
+        });
+    }
+
+    fn update_message(&mut self, styles: &Styles) {
+        let mut msg = progress_bar_msg(&self.stats, self.running, styles);
+        msg += "     ";
+
+        if let Some(running_tests) = &self.running_tests {
+            let (_, width) = console::Term::stderr().size();
+            let width = max(width as usize, 40);
+            let now = Instant::now();
+            let mut count = match self.max_progress_running {
+                MaxProgressRunning::Count(count) => min(running_tests.len(), count),
+                MaxProgressRunning::Infinite => running_tests.len(),
+            };
+            for running_test in &running_tests[..count] {
+                msg.push('\n');
+                msg.push_str(&running_test.message(&now, width, styles));
+            }
+            if count < running_tests.len() {
+                let overflow_count = running_tests.len() - count;
+                msg.push_str(&format!(
+                    "\n             ... and {} more {} running",
+                    overflow_count.style(styles.count),
+                    plural::tests_str(overflow_count),
+                ));
+                count += 1;
+            }
+            self.max_running_displayed = max(self.max_running_displayed, count);
+            msg.push_str(&"\n".to_string().repeat(self.max_running_displayed - count));
+        }
+        self.bar.set_message(msg);
     }
 
     pub(super) fn update_progress_bar(&mut self, event: &TestEvent<'_>, styles: &Styles) {
         let before_should_hide = self.should_hide();
 
         match &event.kind {
+            TestEventKind::StressSubRunStarted { .. } => {
+                self.bar.reset();
+            }
             TestEventKind::StressSubRunFinished { .. } => {
                 // Clear all test bars to remove empty lines of output between
                 // sub-runs.
-                if let Some(test_bars) = &self.test_bars {
-                    test_bars.finish_and_clear_all();
-                }
+                self.bar.finish_and_clear();
                 // Hide the progress bar between sub runs to avoid a spurious
                 // progress bar.
                 self.hidden_between_sub_runs = true;
@@ -306,8 +333,10 @@ impl ProgressBarState {
             TestEventKind::TestStarted {
                 current_stats,
                 running,
+                test_instance,
                 ..
             } => {
+                self.running = *running;
                 self.hidden_between_sub_runs = false;
 
                 self.bar.set_prefix(progress_bar_prefix(
@@ -315,21 +344,19 @@ impl ProgressBarState {
                     current_stats.cancel_reason,
                     styles,
                 ));
-                self.bar
-                    .set_message(progress_bar_msg(current_stats, *running, styles));
                 // If there are skipped tests, the initial run count will be lower than when constructed
                 // in ProgressBar::new.
                 self.bar.set_length(current_stats.initial_run_count as u64);
                 self.bar.set_position(current_stats.finished_count as u64);
-            }
-            TestEventKind::TestShowProgress { test_instance, .. } => {
-                if let Some(test_bars) = &mut self.test_bars {
-                    test_bars.add_test(
-                        &self.multi_progress,
-                        &test_instance.id(),
-                        String::new(),
-                        styles,
-                    );
+
+                if let Some(running_tests) = &mut self.running_tests {
+                    running_tests.push(RunningTest {
+                        binary_id: test_instance.id().binary_id.clone(),
+                        test_name: test_instance.id().test_name.to_owned(),
+                        status: RunningTestStatus::Running,
+                        start_time: Instant::now(),
+                        paused_for: Duration::ZERO,
+                    });
                 }
             }
             TestEventKind::TestFinished {
@@ -338,9 +365,8 @@ impl ProgressBarState {
                 test_instance,
                 ..
             } => {
-                if let Some(test_bars) = &mut self.test_bars {
-                    test_bars.remove_test(&self.multi_progress, &test_instance.id());
-                }
+                self.running = *running;
+                self.remove_test(&test_instance.id());
 
                 self.hidden_between_sub_runs = false;
 
@@ -349,43 +375,49 @@ impl ProgressBarState {
                     current_stats.cancel_reason,
                     styles,
                 ));
-                self.bar
-                    .set_message(progress_bar_msg(current_stats, *running, styles));
                 // If there are skipped tests, the initial run count will be lower than when constructed
                 // in ProgressBar::new.
                 self.bar.set_length(current_stats.initial_run_count as u64);
                 self.bar.set_position(current_stats.finished_count as u64);
             }
-            TestEventKind::TestAttemptFailedWillRetry { test_instance, .. } => {
-                if let Some(test_bars) = &mut self.test_bars {
-                    test_bars.remove_test(&self.multi_progress, &test_instance.id());
-                    // TODO: it would be nice to count the delay down rather
-                    // than up. But it probably will require changes to
-                    // indicatif to enable that (or for us to render time
-                    // ourselves).
-                    test_bars.add_test(
-                        &self.multi_progress,
-                        &test_instance.id(),
-                        "DELAY".style(styles.retry).to_string(),
-                        styles,
-                    );
+            TestEventKind::TestAttemptFailedWillRetry {
+                test_instance,
+                delay_before_next_attempt,
+                ..
+            } => {
+                self.remove_test(&test_instance.id());
+                if let Some(running_tests) = &mut self.running_tests {
+                    running_tests.push(RunningTest {
+                        binary_id: test_instance.id().binary_id.clone(),
+                        test_name: test_instance.id().test_name.to_owned(),
+                        status: RunningTestStatus::Delay(*delay_before_next_attempt),
+                        start_time: Instant::now(),
+                        paused_for: Duration::ZERO,
+                    });
                 }
             }
             TestEventKind::TestRetryStarted { test_instance, .. } => {
-                if let Some(test_bars) = &mut self.test_bars {
-                    test_bars.remove_test(&self.multi_progress, &test_instance.id());
-                    test_bars.add_test(
-                        &self.multi_progress,
-                        &test_instance.id(),
-                        "RETRY".style(styles.retry).to_string(),
-                        styles,
-                    );
+                self.remove_test(&test_instance.id());
+                if let Some(running_tests) = &mut self.running_tests {
+                    running_tests.push(RunningTest {
+                        binary_id: test_instance.id().binary_id.clone(),
+                        test_name: test_instance.id().test_name.to_owned(),
+                        status: RunningTestStatus::Retry,
+                        start_time: Instant::now(),
+                        paused_for: Duration::ZERO,
+                    });
                 }
             }
             TestEventKind::TestSlow { test_instance, .. } => {
-                if let Some(test_bars) = &mut self.test_bars {
-                    test_bars
-                        .set_prefix(&test_instance.id(), "SLOW".style(styles.skip).to_string());
+                if let Some(running_tests) = &mut self.running_tests {
+                    running_tests
+                        .iter_mut()
+                        .find(|rt| {
+                            &rt.binary_id == test_instance.id().binary_id
+                                && rt.test_name == test_instance.id().test_name
+                        })
+                        .expect("a slow test to be already running")
+                        .status = RunningTestStatus::Slow;
                 }
             }
             TestEventKind::InfoStarted { .. } => {
@@ -414,9 +446,11 @@ impl ProgressBarState {
                 // though.
                 self.bar.clone().with_elapsed(event.elapsed);
 
-                let delta = current_global_elapsed.saturating_sub(event.elapsed);
-                if let Some(test_bars) = &mut self.test_bars {
-                    test_bars.on_run_continued(delta);
+                if let Some(running_tests) = &mut self.running_tests {
+                    let delta = current_global_elapsed.saturating_sub(event.elapsed);
+                    for running_test in running_tests {
+                        running_test.paused_for += delta;
+                    }
                 }
             }
             TestEventKind::RunBeginCancel { current_stats, .. }
@@ -432,32 +466,38 @@ impl ProgressBarState {
         let after_should_hide = self.should_hide();
 
         match (before_should_hide, after_should_hide) {
-            (false, true) => self.multi_progress.set_draw_target(Self::hidden_target()),
-            (true, false) => self.multi_progress.set_draw_target(Self::stderr_target()),
+            (false, true) => self.bar.set_draw_target(Self::hidden_target()),
+            (true, false) => self.bar.set_draw_target(Self::stderr_target()),
             _ => {}
         }
     }
 
-    pub(super) fn write_buf(&self, buf: &[u8]) -> io::Result<()> {
-        // ProgressBar::println doesn't print status lines if the bar is
-        // hidden. The suspend method prints it in all cases.
-        self.multi_progress
-            .suspend(|| std::io::stderr().write_all(buf))
+    fn remove_test(&mut self, test_instance: &TestInstanceId) {
+        if let Some(running_tests) = &mut self.running_tests {
+            running_tests.remove(
+                running_tests
+                    .iter()
+                    .position(|e| {
+                        &e.binary_id == test_instance.binary_id
+                            && e.test_name == test_instance.test_name
+                    })
+                    .expect("finished test to have started"),
+            );
+        }
+    }
+
+    pub(super) fn write_buf(&mut self, buf: &str) {
+        self.buffer.push_str(buf);
     }
 
     #[inline]
     pub(super) fn finish_and_clear(&self) {
+        self.print_and_force_redraw();
         self.bar.finish_and_clear();
-        if let Some(test_bars) = &self.test_bars {
-            test_bars.finish_and_clear_all();
-        }
     }
 
     fn stderr_target() -> ProgressDrawTarget {
-        // This used to be unbuffered, but that option went away from indicatif
-        // 0.17.0. The refresh rate is now 20hz so that it's double the steady
-        // tick rate.
-        ProgressDrawTarget::stderr_with_hz(20)
+        ProgressDrawTarget::stderr_with_hz(PROGRESS_REFRESH_RATE_HZ)
     }
 
     fn hidden_target() -> ProgressDrawTarget {
@@ -472,12 +512,15 @@ impl ProgressBarState {
     }
 
     pub(super) fn is_hidden(&self) -> bool {
-        self.multi_progress.is_hidden()
+        self.bar.is_hidden()
     }
 }
 
 /// OSC 9 terminal progress reporting.
-pub(super) struct TerminalProgress {}
+#[derive(Default)]
+pub(super) struct TerminalProgress {
+    last_value: TerminalProgressValue,
+}
 
 impl TerminalProgress {
     const ENV: &str = "CARGO_TERM_PROGRESS_TERM_INTEGRATION";
@@ -490,7 +533,7 @@ impl TerminalProgress {
                     if let Some(v) = config.term.progress.term_integration {
                         if v {
                             debug!("enabling terminal progress reporting based on {source:?}");
-                            return Some(Self {});
+                            return Some(Self::default());
                         } else {
                             debug!("disabling terminal progress reporting based on {source:?}");
                             return None;
@@ -504,7 +547,7 @@ impl TerminalProgress {
                                 "enabling terminal progress reporting based on \
                                  CARGO_TERM_PROGRESS_TERM_INTEGRATION environment variable"
                             );
-                            return Some(Self {});
+                            return Some(Self::default());
                         } else if v == "false" {
                             debug!(
                                 "disabling terminal progress reporting based on \
@@ -523,7 +566,7 @@ impl TerminalProgress {
                     if let Some(v) = config.term.progress.term_integration {
                         if v {
                             debug!("enabling terminal progress reporting based on {source:?}");
-                            return Some(Self {});
+                            return Some(Self::default());
                         } else {
                             debug!("disabling terminal progress reporting based on {source:?}");
                             return None;
@@ -533,14 +576,10 @@ impl TerminalProgress {
             }
         }
 
-        supports_osc_9_4(stream).then_some(TerminalProgress {})
+        supports_osc_9_4(stream).then(Self::default)
     }
 
-    pub(super) fn update_progress(
-        &self,
-        event: &TestEvent<'_>,
-        writer: &mut dyn Write,
-    ) -> Result<(), io::Error> {
+    pub(super) fn update_progress(&mut self, event: &TestEvent<'_>) {
         let value = match &event.kind {
             TestEventKind::RunStarted { .. }
             | TestEventKind::StressSubRunStarted { .. }
@@ -559,8 +598,7 @@ impl TerminalProgress {
                     TerminalProgressValue::Value(percentage)
                 }
             }
-            TestEventKind::TestShowProgress { .. }
-            | TestEventKind::TestSlow { .. }
+            TestEventKind::TestSlow { .. }
             | TestEventKind::TestAttemptFailedWillRetry { .. }
             | TestEventKind::TestRetryStarted { .. }
             | TestEventKind::TestSkipped { .. }
@@ -585,7 +623,11 @@ impl TerminalProgress {
             }
         };
 
-        write!(writer, "{value}")
+        self.last_value = value;
+    }
+
+    pub(super) fn last_value(&self) -> &TerminalProgressValue {
+        &self.last_value
     }
 }
 
@@ -619,9 +661,10 @@ fn supports_osc_9_4(stream: &dyn IsTerminal) -> bool {
 /// A progress status value printable as an ANSI OSC 9;4 escape code.
 ///
 /// Adapted from Cargo 1.87.
-#[derive(PartialEq, Debug)]
-enum TerminalProgressValue {
+#[derive(PartialEq, Debug, Default)]
+pub(super) enum TerminalProgressValue {
     /// No output.
+    #[default]
     None,
     /// Remove progress.
     Remove,
@@ -787,6 +830,7 @@ fn progress_bar_cancel_prefix(reason: Option<CancelReason>, styles: &Styles) ->
         | Some(CancelReason::TestFailure)
         | Some(CancelReason::ReportError)
         | Some(CancelReason::GlobalTimeout)
+        | Some(CancelReason::TestFailureImmediate)
         | Some(CancelReason::Signal)
         | Some(CancelReason::Interrupt)
         | None => "Cancelling",
